/*

  OK, now let's use a "deep" net.

  Besides having *more* layers making the net deeper, we also use different layers.

  In the definition below there are convolutional layers (conv) and maxpool layers (maxpool).

  Convolutional layers are specifically used for image recognition and the way they
  are used here become popular with Yann LeCun' LeNet architecture:

  https://en.wikipedia.org/wiki/LeNet

  The math of convolutional layers is somewhat more complex. A good explanation is given
  in chapter 9 of this book:

  https://www.deeplearningbook.org

  LeNet evolved to complexer and deeper networks (AlexNet, GoogLeNet, ResNet, ecc...)
  bringing all the breakthroughs in machine learning in the last decade.

  For our example I defined a sort of "pimped" LeNet (sorry ;) with a few more layers.

  This network already reaches 89.84 % accuracy even with a limit of just 1 epoch!

  Just a single epoch takes about 2 minutes on my notebook....

  We can run this on a Machine with a Nvidia Tesla V100 using CUDA/CuDNN...

  There it takes just ~ 30 seconds to converge with an accuracy of 95 % :-)

  The trick is that dlib on V100 hardware can make use of reduced precision floatin point
  numbers (marketing speak is "tensorfloat") where it reaches theoretical 125 TFLOPS
  vs. my notebook's real world 0.5 TFLOPS with standard FP32 numbers.


*/


#include <iostream>
#include <dlib/matrix.h>
#include <dlib/dnn.h>
#include "load_images.h"

using std::cout;
using namespace dlib;

int main() {

    // --- load test data as prepared by 01-preprocess-images --------------------------------

    std::vector<matrix<uint8_t>>      train_image;
    std::vector<unsigned long>        train_label;
    std::vector<matrix<uint8_t>>      test_image;
    std::vector<unsigned long>        test_label;

    const int TRAIN_COUNT = 39209;
    load_images("../train_image.bin", "../train_label.bin", TRAIN_COUNT, train_image, train_label);

    const int TEST_COUNT = 12630;
    load_images("../test_image.bin", "../test_label.bin", TEST_COUNT, test_image, test_label);


    // --- define the net and train it -------------------------------------------------------

    using net_type = loss_multiclass_log<fc<43,
                     relu<fc<90,
                     relu<fc<120,
                     max_pool<2,2,2,2,
                     relu<con<32,5,5,1,1,
                     max_pool<2,2,2,2,
                     relu<con<12,5,5,1,1,
                     relu<con<6,3,3,1,1,
                     relu<con<6,3,3,1,1,
                     relu<con<6,3,3,1,1,
                     input<matrix<unsigned char>>
                     >>>>>>>>>>>> >> >> >>;

    net_type net;

    dnn_trainer<net_type> trainer(net);
    trainer.set_learning_rate(0.01);
    trainer.set_min_learning_rate(0.00001);
    trainer.set_mini_batch_size(128);
    trainer.be_verbose();

    // artificially limit training to 1 epochs
    
    trainer.set_max_num_epochs(1);

    if (trainer.get_max_num_epochs() < 10000) {
        cout << "WARNING, max epochs hat been lowered artificially, this is not dlib's default\n";
    }

    trainer.train(train_image, train_label);
    net.clean();

    // --- score the net using the test images -----------------------------------------------

    std::vector<unsigned long> estimates = net(test_image);
    int num_right = 0;
    int num_wrong = 0;
    for (int i = 0; i < test_image.size(); ++i) {
        if (estimates[i] == test_label[i]) {
            num_right++;
        } else {
            num_wrong++;
        }
    }
    cout << "test correct:  " << num_right << "\n";
    cout << "test wrong:    " << num_wrong << "\n";
    cout << "test accuracy: " << num_right / (double)(num_right + num_wrong) << "\n";

    return 0;
}
